package guru.kake.ronove

import com.deepl.api.DeepLException
import com.deepl.api.Translator
import com.google.cloud.translate.Translate
import com.google.cloud.translate.TranslateException
import com.google.cloud.translate.TranslateOptions
import com.google.cloud.translate.Translation
import javafx.beans.property.SimpleDoubleProperty
import javafx.beans.property.SimpleStringProperty
import tornadofx.getValue
import tornadofx.setValue

/**
 * Provides access to different machine translation services through a unified interface
 */
class TranslationBackend {
    companion object {
        private var service: TranslationService? = null
        var serviceName = SimpleStringProperty()
        var quotaMax = 0L
        var quotaUsed = 0L
        val quotaProperty = SimpleDoubleProperty()
        var quota by quotaProperty
        var destinationLanguage: String = ""
        var sourceLanguage: String = ""

        /**
         * Translate one string of text
         * @param text Text to translate
         * @param sourceLang Override source language, otherwise defaults to what is set in [TranslationBackend]
         * @param destLang Override destination language, otherwise defaults to what is set in [TranslationBackend]
         * @return Translated string
         */
        operator fun invoke(text: String, sourceLang: String = sourceLanguage, destLang: String = destinationLanguage)
            : String = service!!.translate(text, sourceLang, destLang)

        /**
         * Translate a list of strings
         * @param textList List to translate
         * @param sourceLang Override source language, otherwise defaults to what is set in [TranslationBackend]
         * @param destLang Override destination language, otherwise defaults to what is set in [TranslationBackend]
         * @return List of translated strings
         * @throws TranslationServiceUnitializedException When no translation service has been initialized
         */
        operator fun invoke(textList: List<String>, sourceLang: String = sourceLanguage, destLang: String = destinationLanguage)
            : List<String> = service?.translate(textList, sourceLang, destLang) ?:
            throw TranslationServiceUnitializedException("Translation service uninitialized!")

        val provider: String = when (service) {
            null -> "uninitialized"
            is DeepLTranslator -> "DeepL"
            else -> "error"
        }

        fun setService(service: TranslationService, name: String) {
            this.service = service
            serviceName.value = name
        }
    }
}

/**
 * Interface for machine translating strings and lists of strings
 */
interface TranslationService {
    /**
     * Translate a single String of text
     *
     * @param text Text to translate
     * @param sourceLang Override source language, otherwise defaults to what is set in [TranslationBackend]
     * @param destLang Override destination language, otherwise defaults to what is set in [TranslationBackend]
     * @return Translated string
     */
    fun translate(text: String, sourceLang: String, destLang: String): String

    /**
     * Translate multiple Strings of text.
     *
     * Default implementation calls the single string translate() repeatedly, if the backed supports a way to translate
     * multiple strings in one request, override this function to do that instead.
     *
     * @param textList List to translate
     * @param sourceLang Override source language, otherwise defaults to what is set in [TranslationBackend]
     * @param destLang Override destination language, otherwise defaults to what is set in [TranslationBackend]
     * @return List of translated strings
     * @throws TranslationServiceUnitializedException When no translation service has been initialized
     */
    fun translate(textList: List<String>, sourceLang: String, destLang: String): List<String> =
        List(textList.size) {
            translate(textList[it], sourceLang, destLang)
        }
}

/**
 * TranslationsService implementation using DeepL online translation
 *
 * @param[key] DeepL API key
 * @throws[TranslationServiceException] wraps API-specific exceptions
 */
class DeepLTranslator(key: String) : TranslationService {
    private var deepl: Translator

    init {
        try {
            deepl = Translator(key)
            val usage = deepl.usage
            TranslationBackend.quotaUsed = usage.character!!.count
            TranslationBackend.quotaMax = usage.character!!.limit
            TranslationBackend.quota = usage.character!!.count.toDouble() / usage.character!!.limit
        } catch (e: DeepLException) {
            throw TranslationServiceException("Error: ${e.message}", e)
        } catch (e: IllegalArgumentException) {
            throw TranslationServiceException("API key must not be empty")
        }
    }

    override fun translate(text: String, sourceLang: String, destLang: String): String {
        try {
            return deepl.translateText(text, sourceLang, destLang).text
        } catch (e: DeepLException) {
            throw TranslationServiceException("DeepL API translation failed with message: ${e.message}", e)
        }
    }

    override fun translate(textList: List<String>, sourceLang: String, destLang: String): List<String> {
        try {
            val translations = deepl.translateText(textList, sourceLang, destLang)
            return List<String>(translations.size) { translations[it].text }
        } catch (e: DeepLException) {
            throw TranslationServiceException("DeepL API translation failed with message: ${e.message}", e)
        }
    }
}

/**
 * TranslationsService implementation using Google Translate online translation using the basic, free v2 API
 *
 * @param[key] Google Cloud Translate API key
 * @throws[TranslationServiceException] wraps API-specific exceptions
 */
class GoogleTranslator(key: String) : TranslationService {
    private val translate: Translate
    init {
        @Suppress("DEPRECATION")
        translate = TranslateOptions.newBuilder().setApiKey(key).build().service
    }

    override fun translate(text: String, sourceLang: String, destLang: String): String {
        try {
            val translation: Translation = translate.translate(
                text,
                Translate.TranslateOption.sourceLanguage(sourceLang),
                Translate.TranslateOption.targetLanguage(destLang)
            )
            return translation.translatedText
        } catch (e: TranslateException) {
            throw TranslationServiceException("Google API translation failed with message: ${e.message}", e)
        }
    }

    override fun translate(textList: List<String>, sourceLang: String, destLang: String): List<String> {
        try {
            val translations: List<Translation> = translate.translate(
                textList,
                Translate.TranslateOption.sourceLanguage(sourceLang),
                Translate.TranslateOption.targetLanguage(destLang)
            )
            return List<String>(translations.size) { translations[it].translatedText }
        } catch (e: TranslateException) {
            throw TranslationServiceException("Google API translation failed with message: ${e.message}", e)
        }
    }
}

class TranslationServiceException(message: String? = null, cause: Throwable? = null) : Exception(message, cause)
class TranslationServiceUnitializedException(message: String? = null) : Exception(message)